package net.knarcraft.blacksmithvisuals.listener;

import com.comphenix.protocol.PacketType;
import com.comphenix.protocol.ProtocolLibrary;
import com.comphenix.protocol.ProtocolManager;
import com.comphenix.protocol.events.PacketContainer;
import net.citizensnpcs.api.npc.NPC;
import net.citizensnpcs.trait.LookClose;
import net.knarcraft.blacksmith.event.ActionStartEvent;
import net.knarcraft.blacksmith.event.BlacksmithReforgeFailEvent;
import net.knarcraft.blacksmith.event.BlacksmithReforgeStartEvent;
import net.knarcraft.blacksmith.event.BlacksmithReforgeSucceedEvent;
import net.knarcraft.blacksmith.event.NPCSoundEvent;
import net.knarcraft.blacksmith.event.ScrapperSalvageFailEvent;
import net.knarcraft.blacksmith.event.ScrapperSalvageStartEvent;
import net.knarcraft.blacksmith.event.ScrapperSalvageSucceedEvent;
import net.knarcraft.blacksmithvisuals.BlacksmithVisuals;
import net.knarcraft.blacksmithvisuals.container.AnimationData;
import net.knarcraft.blacksmithvisuals.container.NPCData;
import net.knarcraft.blacksmithvisuals.container.SoundData;
import net.knarcraft.blacksmithvisuals.manager.ConfigurationManager;
import net.knarcraft.blacksmithvisuals.property.NPCPosition;
import net.knarcraft.blacksmithvisuals.property.SoundIdentifier;
import net.knarcraft.blacksmithvisuals.util.SoundHelper;
import net.knarcraft.knarlib.particle.ParticleConfig;
import net.knarcraft.knarlib.particle.ParticleMode;
import net.knarcraft.knarlib.util.ParticleHelper;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.Particle;
import org.bukkit.entity.Entity;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.event.player.PlayerTeleportEvent;
import org.bukkit.scheduler.BukkitScheduler;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.HashMap;
import java.util.Map;
import java.util.Random;
import java.util.UUID;
import java.util.logging.Level;

/**
 * A listener for blacksmith-related events
 */
public class BlacksmithListener implements Listener {

    private final Random random;
    private final ConfigurationManager configurationManager;
    private final Map<UUID, Double> oldLookRanges;

    /**
     * Instantiates a new blacksmith listener
     *
     * @param configurationManager <p>The configuration manager to get configuration values from</p>
     */
    public BlacksmithListener(@NotNull ConfigurationManager configurationManager) {
        this.random = new Random();
        this.configurationManager = configurationManager;
        this.oldLookRanges = new HashMap<>();
    }

    @EventHandler
    public void onDefaultSound(@NotNull NPCSoundEvent event) {
        BlacksmithVisuals.debug("Detected onDefaultSound event");
        event.setCancelled(true);
    }

    @EventHandler
    public void onReforgeStart(@NotNull BlacksmithReforgeStartEvent event) {
        BlacksmithVisuals.debug("Detected onReforgeStart event");
        runWorkingAnimation(event, SoundIdentifier.REFORGING_WORKING, this.configurationManager.getBlacksmithAnimationData());
    }

    @EventHandler
    public void onSalvageStart(@NotNull ScrapperSalvageStartEvent event) {
        BlacksmithVisuals.debug("Detected onSalvageStart event");
        runWorkingAnimation(event, SoundIdentifier.SALVAGING_WORKING, this.configurationManager.getScrapperAnimationData());
    }

    @EventHandler
    public void onReforgeSuccess(@NotNull BlacksmithReforgeSucceedEvent event) {
        BlacksmithVisuals.debug("Detected onReforgeSuccess event");
        SoundHelper.playSound(event.getEntity(), this.configurationManager.getSoundData(SoundIdentifier.REFORGING_SUCCESS));
    }

    @EventHandler
    public void onSalvageSuccess(@NotNull ScrapperSalvageSucceedEvent event) {
        BlacksmithVisuals.debug("Detected onSalvageSuccess event");
        SoundHelper.playSound(event.getEntity(), this.configurationManager.getSoundData(SoundIdentifier.SALVAGING_SUCCESS));
    }

    @EventHandler
    public void onReforgeFail(@NotNull BlacksmithReforgeFailEvent event) {
        BlacksmithVisuals.debug("Detected onReforgeFail event");
        SoundHelper.playSound(event.getEntity(), this.configurationManager.getSoundData(SoundIdentifier.REFORGING_FAILURE));
    }

    @EventHandler
    public void onSalvageFail(@NotNull ScrapperSalvageFailEvent event) {
        BlacksmithVisuals.debug("Detected onSalvageFail event");
        SoundHelper.playSound(event.getEntity(), this.configurationManager.getSoundData(SoundIdentifier.SALVAGING_FAILURE));
    }

    /**
     * Runs the working animation for an NPC
     *
     * @param event           <p>The action that started</p>
     * @param soundIdentifier <p>The identifier for the sound to play</p>
     * @param animationData   <p>The animation data for the animation to play</p>
     */
    private void runWorkingAnimation(@NotNull ActionStartEvent event, @NotNull SoundIdentifier soundIdentifier,
                                     @NotNull AnimationData animationData) {
        BlacksmithVisuals.debug("Preparing to run working animation");
        BlacksmithVisuals instance = BlacksmithVisuals.getInstance();
        BukkitScheduler scheduler = Bukkit.getScheduler();
        NPC npc = event.getNpc();

        NPCPosition npcPosition = NPCPosition.getFromMaterial(event.getCraftingStation());
        BlacksmithVisuals.debug("Found NPC position: " + npcPosition);
        long delay = moveToWorkingPosition(npc, event.getEntity(), npcPosition);
        BlacksmithVisuals.debug("Calculated delay to: " + delay);
        if (npc.hasTrait(LookClose.class)) {
            BlacksmithVisuals.debug("Found LookClose trait");
            LookClose trait = npc.getTraitNullable(LookClose.class);
            this.oldLookRanges.put(npc.getUniqueId(), trait.getRange());
            trait.setRange(0);
            BlacksmithVisuals.debug("Edited LookClose range");
        }
        long finishTime = event.getActionDurationTicks() - (2 * delay);

        BlacksmithVisuals.debug("Scheduling work animation and back movement");
        scheduler.scheduleSyncDelayedTask(instance, () -> startWorkAnimation(event.getEntity(),
                this.configurationManager.getSoundData(soundIdentifier), animationData, finishTime - 20, npcPosition), delay);
        scheduler.scheduleSyncDelayedTask(instance, () -> moveBack(event.getNpc(), event.getEntity()), finishTime);
    }

    /**
     * Moves an NPC back to its idle position
     *
     * @param npc    <p>The NPC to move</p>
     * @param entity <p>The entity of the NPC</p>
     */
    private void moveBack(@NotNull NPC npc, @NotNull Entity entity) {
        if (!npc.isSpawned()) {
            BlacksmithVisuals.debug("NPC isn't spawned, and thus cannot be moved back");
            return;
        }

        NPCData npcData = BlacksmithVisuals.getInstance().getNpcDataManager().getData(npc.getUniqueId());
        if (npcData == null) {
            BlacksmithVisuals.debug("Unable to get NPC data for NPC");
            return;
        }

        Location targetLocation = npcData.positions().get(NPCPosition.IDLE);
        if (!npc.getNavigator().canNavigateTo(targetLocation)) {
            BlacksmithVisuals.getInstance().getLogger().log(Level.WARNING, "Idle position for NPC " +
                    npc.getName() + " is unreachable!");
            return;
        }

        if (npc.hasTrait(LookClose.class)) {
            LookClose trait = npc.getTraitNullable(LookClose.class);
            Double oldRange = oldLookRanges.remove(npc.getUniqueId());
            if (oldRange != null) {
                trait.setRange(oldRange);
            }
            BlacksmithVisuals.debug("Restored LookCLose range");
        }

        moveNPC(npc, entity, targetLocation);
    }

    /**
     * Moves a npc to its working position
     *
     * @param npc         <p>The NPC to move</p>
     * @param entity      <p>The entity of the NPC</p>
     * @param npcPosition <p>The npc position to move to</p>
     * @return <p>The time the move will take</p>
     */
    private long moveToWorkingPosition(@NotNull NPC npc, @NotNull Entity entity, @Nullable NPCPosition npcPosition) {
        if (!npc.isSpawned() || npcPosition == null) {
            BlacksmithVisuals.debug("The NPC isn't spawned, or its working position doesn't exist");
            return 0;
        }

        NPCData npcData = BlacksmithVisuals.getInstance().getNpcDataManager().getData(npc.getUniqueId());
        if (npcData == null) {
            BlacksmithVisuals.debug("NPC sata could not be retrieved");
            return 0;
        }

        Location targetLocation = npcData.positions().get(npcPosition);
        if (targetLocation == null && npcPosition != NPCPosition.WORKING_REPAIRABLE) {
            targetLocation = npcData.positions().get(NPCPosition.WORKING_REPAIRABLE);
        }
        if (targetLocation == null) {
            BlacksmithVisuals.debug("Working location not set for WORKING_REPAIRABLE");
            return 0;
        }

        if (!npc.getNavigator().canNavigateTo(targetLocation)) {
            BlacksmithVisuals.getInstance().getLogger().log(Level.WARNING, "Working position for NPC " +
                    npc.getName() + " is unreachable!");
            return 0;
        }

        return moveNPC(npc, entity, targetLocation);
    }

    /**
     * Moves an NPC to a different location
     *
     * @param npc                 <p>The NPC to be moved</p>
     * @param entity              <p>The NPC's entity</p>
     * @param finalTargetLocation <p>The location the NPC should be moved to</p>
     * @return <p>The assumed time it will take the NPC to get to the position</p>
     */
    private long moveNPC(@NotNull NPC npc, @NotNull Entity entity, @NotNull Location finalTargetLocation) {
        // Move NPC using Citizens path-finding
        BlacksmithVisuals.debug("Preparing rotation and walk to working location");
        Location current = entity.getLocation().clone();
        npc.teleport(current.setDirection(finalTargetLocation.toVector().subtract(current.toVector())), PlayerTeleportEvent.TeleportCause.PLUGIN);
        Bukkit.getScheduler().scheduleSyncDelayedTask(BlacksmithVisuals.getInstance(), () -> npc.getNavigator().setTarget(finalTargetLocation), 2);

        // Teleport the NPC tp get it in the exact final location
        long walkTime = getWalkTime(entity.getLocation(), finalTargetLocation);
        BlacksmithVisuals.debug("Queuing teleportation to working position");
        Bukkit.getScheduler().scheduleSyncDelayedTask(BlacksmithVisuals.getInstance(), () ->
                entity.teleport(finalTargetLocation, PlayerTeleportEvent.TeleportCause.PLUGIN), walkTime);
        return walkTime;
    }

    /**
     * Starts the working animation and sound
     *
     * @param entity        <p>The entity to play the sound at</p>
     * @param soundData     <p>The sound data for the sound to play</p>
     * @param animationData <p>The animation data for the animation to play</p>
     * @param durationTicks <p>The duration of the work animation</p>
     * @param npcPosition   <p>The npc position to animate</p>
     */
    private void startWorkAnimation(@NotNull Entity entity, @NotNull SoundData soundData,
                                    @NotNull AnimationData animationData, long durationTicks,
                                    @Nullable NPCPosition npcPosition) {
        BlacksmithVisuals.debug("Starting work animation");
        BlacksmithVisuals instance = BlacksmithVisuals.getInstance();
        BukkitScheduler scheduler = Bukkit.getScheduler();

        BlacksmithVisuals.debug("Scheduling animation");
        int playWorkSound = scheduler.scheduleSyncRepeatingTask(instance,
                () -> animateNPC(entity, soundData, animationData, npcPosition), 20, animationData.animationDelay());
        scheduler.scheduleSyncDelayedTask(instance, () -> scheduler.cancelTask(playWorkSound), durationTicks);
        BlacksmithVisuals.debug("Stopped working animation");
    }

    /**
     * Gets the time it would take to go from one location to another in ticks
     *
     * @param startLocation  <p>The location to start from</p>
     * @param targetLocation <p>The target location</p>
     * @return <p>The time in ticks</p>
     */
    private long getWalkTime(@NotNull Location startLocation, @NotNull Location targetLocation) {
        double distance = startLocation.distance(targetLocation);
        double WALK_SPEED = 4.317;
        return Math.round((distance / WALK_SPEED) * 20);
    }

    /**
     * Animates a npc's hand, and plays a sound
     *
     * @param entity        <p>The entity to animate and play the sound at</p>
     * @param soundData     <p>The sound to be played</p>
     * @param animationData <p>The animation to be played</p>
     * @param npcPosition   <p>The npc position to animate</p>
     */
    private void animateNPC(@NotNull Entity entity, @NotNull SoundData soundData, @NotNull AnimationData animationData,
                            @Nullable NPCPosition npcPosition) {
        // Don't play disabled animations
        if (!animationData.animateOffHand()) {
            BlacksmithVisuals.debug("Offhand animation is disabled");
            return;
        }

        if (random.nextInt(100) >= animationData.animationChance()) {
            BlacksmithVisuals.debug("Animation skipped because of RNG");
            return;
        }

        BlacksmithVisuals.debug("Playing work sound");
        SoundHelper.playSound(entity, soundData);

        if (soundData.offsetTicks() < 0) {
            BlacksmithVisuals.debug("Delaying off-hand animation");
            Bukkit.getScheduler().scheduleSyncDelayedTask(BlacksmithVisuals.getInstance(),
                    () -> animateOffhand(entity), -soundData.offsetTicks());
        } else {
            BlacksmithVisuals.debug("Animating off-hand");
            animateOffhand(entity);
        }

        if (npcPosition == null) {
            npcPosition = NPCPosition.WORKING_REPAIRABLE;
        }
        for (int i = 0; i < random.nextInt(5) + 1; i++) {
            BlacksmithVisuals.debug("Spawning particle");
            spawnSparkParticle(entity, npcPosition);
        }
    }

    /**
     * Spawns a spark particle at the given NPC's work station
     *
     * @param entity      <p>The entity whose working station should produce a spark</p>
     * @param npcPosition <p>The npc potion to spawn a particle for</p>
     */
    private void spawnSparkParticle(@NotNull Entity entity, @NotNull NPCPosition npcPosition) {
        BlacksmithVisuals.debug("Preparing particle spawning");
        double randomX = this.random.nextDouble(-1, 1);
        double randomY = this.random.nextDouble(0.1, 1);
        double randomZ = this.random.nextDouble(-1, 1);
        Particle particle;
        if (npcPosition == NPCPosition.WORKING_CRAFTING) {
            particle = Particle.ELECTRIC_SPARK;
        } else {
            particle = Particle.FLAME;
        }
        Location spawnLocation = entity.getLocation().clone().getBlock().getRelative(entity.getFacing()).getLocation().add(0.5, 0, 0.5);
        BlacksmithVisuals.debug("Calculated particle location to " + spawnLocation);
        ParticleConfig particleConfig = new ParticleConfig(ParticleMode.SINGLE, particle, 0, 0, 1.1, randomX, randomY, randomZ, 0.1);
        ParticleHelper.spawnParticle(entity.getWorld(), spawnLocation, particleConfig, 1);
        BlacksmithVisuals.debug("Spawned particle");
    }

    /**
     * Animates an NPC's offhand
     *
     * @param entity <p>The entity to animate</p>
     */
    private void animateOffhand(@NotNull Entity entity) {
        BlacksmithVisuals.debug("Preparing off-hand animation");
        ProtocolManager manager = ProtocolLibrary.getProtocolManager();
        PacketContainer packet = manager.createPacket(PacketType.Play.Server.ANIMATION);
        packet.getIntegers().write(0, entity.getEntityId());
        packet.getIntegers().write(1, 3);

        for (Player player : Bukkit.getOnlinePlayers()) {
            if (player.getWorld().equals(entity.getWorld())) {
                BlacksmithVisuals.debug("Sending animation packet to " + player);
                manager.sendServerPacket(player, packet);
            }
        }
        BlacksmithVisuals.debug("Finished sending animation packets");
    }


}
